from collections.abc import Collection, Callable, Sequence
from dataclasses import dataclass
from functools import partial
from re import Pattern
from typing import TypeVar, Protocol

from token_distance.algorithms.token_matching import match_tokens
from token_distance.means import Mean, average, get_mean_by_name
from token_distance.similarities import Similarity, jaro_winkler, get_similarity_by_name
from token_distance.types.registry import get_if_needed
from token_distance.types.token import WeightedToken, Token, EvaluatedToken, split_text_by, parse_token
from token_distance.weights import Weight, even_weight, get_weight_by_name


def token_distance(
    similarity: Similarity,
    mean: Mean,
    weight: Weight,
    tokens: Collection[Token],
    candidates: Collection[str]
) -> float:
    """
    Calculate the token distance between a collection of tokens and a collection of candidates.

    Args:
        similarity (Similarity): A function that calculates the similarity between two tokens.
        mean (Mean): A function that calculates the mean value of a collection of values.
        weight (Weight): A function that assigns weights to tokens.
        tokens (Collection[Token]): A collection of tokens.
        candidates (Collection[str]): A collection of candidate strings.

    Returns:
        float: The token distance between the tokens and candidates.
    """
    weighted_tokens: Collection[WeightedToken] = weight(tokens)
    evaluated_tokens: Collection[EvaluatedToken] = [
        max(similarity(token, candidate) for candidate in candidates)
        for token in weighted_tokens
    ]
    return mean(evaluated_tokens)




def exclusive_token_distance(
    similarity: Similarity,
    mean: Mean,
    weight: Weight,
    tokens: Collection[Token],
    candidates: Collection[str]
) -> float:
    """
    Calculate the token distance between a collection of tokens and a collection of candidates.
    If a token gets matched to a candidate, this candidate will no longer be matchable for other tokens,
    if not multiple instances of candidate exist in the text.

    Args:
        similarity: A Similarity instance used for calculating token similarity.
        mean: A Mean instance used for calculating the mean of evaluated tokens.
        weight: A Weight instance used for weighting tokens.
        tokens: A collection of Token instances representing the tokens to be evaluated.
        candidates: A collection of strings representing the candidate tokens to compare with.

    Returns:
        float: The result of applying the mean function to the evaluated tokens.

    """
    return mean(match_tokens(similarity, weight, True, tokens, candidates))


@dataclass
class Config:
    """
    Config represents the configuration options for token distance.

    Attributes:
        similarity (Similarity | str): The similarity metric to be used. Defaults to 'jaro_winkler'.
        mean (Mean | str): The mean calculation method to be used. Defaults to 'average'.
        weight (Weight | str): The weighting method to be used. Defaults to 'even_weight'.
        tokenize (Callable[[str], Sequence[str]] | Pattern | str | None): The tokenization method to be used.
            Defaults to None.
        exclusive (bool): Whether exclusive matching should be used. Defaults to False.
    """
    similarity: Similarity | str = jaro_winkler
    mean: Mean | str = average
    weight: Weight | str = even_weight
    tokenize: Callable[[str], Sequence[str]] | Pattern | str | None = None
    exclusive: bool = False


_T = TypeVar('_T')





class TokenDistanceAlgorithm(Protocol):
    """
    This class represents a token distance algorithm.

    Methods:
    - __call__: Calculates the token distance between two strings.

    """
    def __call__(self, left: str, right: str) -> float:
        ...



def from_config(config: Config = Config()) -> TokenDistanceAlgorithm:
    """
    Constructs a TokenDistanceAlgorithm function based on the provided configuration.

    Args:
        config: A Config object containing the algorithm configuration.

    Returns:
        A TokenDistanceAlgorithm function that calculates the distance between two strings.
    """
    base_algorithm: Callable[
        [Similarity, Mean, Weight, Collection[Token], Collection[str]],
        float
    ] = exclusive_token_distance if config.exclusive else token_distance

    algorithm: Callable[[Collection[Token], Collection[str]], float] = partial(
        base_algorithm,
        get_if_needed(get_similarity_by_name, config.similarity),
        get_if_needed(get_mean_by_name, config.mean),
        get_if_needed(get_weight_by_name, config.weight),
    )

    def distance(left: str, right: str) -> float:
        """
        Args:
            left: A string representing the left text.
            right: A string representing the right text.

        Returns:
            A float representing the token distance between the left and right texts.
        """
        return algorithm(
            tuple(map(parse_token, split_text_by(left, config.tokenize))),
            split_text_by(right, config.tokenize)
        )

    return distance






from collections.abc import Iterator, Callable
from typing import TypeVar, Generic, Sequence, runtime_checkable, Protocol


@runtime_checkable
class HasName(Protocol):
    __name__: str


_T = TypeVar('_T', bound=HasName)


class AlgorithmRegistry(Generic[_T]):
    """
    Class representing a registry for algorithms.

    This class is used to store and retrieve algorithms by their names. It provides functionality to add algorithms to
    the registry and retrieve them using indexing.

    Attributes:
        _algorithms (dict[str, _T]): A dictionary to store the algorithms with their names as keys.

    Methods:
        __init__(*algorithms: _T): Initializes a new AlgorithmRegistry instance with the specified algorithms.
        supported() -> Sequence[str]: Returns a sorted sequence of names of supported algorithms.
        __getitem__(name: str) -> _T: Retrieves the algorithm with the specified name from the registry.

    Example usage:
        # Create an AlgorithmRegistry instance with two algorithms
        registry = AlgorithmRegistry(algorithm1, algorithm2)

        # Get the names of supported algorithms
        supported_algorithms = registry.supported()

        # Retrieve an algorithm by its name
        algorithm = registry['algorithm1']
    """
    def __init__(self, *algorithms: _T):
        self._algorithms: dict[str, _T] = dict(
            (algo.__name__, algo)
            for algo in algorithms
        )

    def __iter__(self) -> Iterator[_T]:
        return iter(self._algorithms.values())

    @property
    def supported(self) -> Sequence[str]:
        """
        Returns a sorted sequence of supported algorithms.

        Returns:
            Sequence[str]: A sorted sequence of supported algorithms.
        """
        return sorted(self._algorithms)

    def __getitem__(self, name: str) -> _T:
        try:
            return self._algorithms[name]
        except KeyError:
            raise NameError(f'{name} not known. Supported names are {", ".join(self.supported)}')


def get_if_needed(get: Callable[[str], _T], value: _T | str) -> _T:
    """
    Args:
        get: A callable function that takes a string as input and returns a value.
        value: Either a value of type _T or a string. If it is a string, it will be passed to the "get" function to
        retrieve the corresponding value.

    Returns:
        A value of type _T. If the "value" parameter is not a string, it is returned as is. If it is a string, the
        "get" function is called with the string as input to retrieve the corresponding value, which is then returned.

    """
    return value if not isinstance(value, str) else get(value)


from collections.abc import Callable, Collection, Sequence
from dataclasses import replace, dataclass
from functools import partial
from operator import attrgetter
from re import Pattern
from typing import Protocol

from token_distance.similarities import RecordingSimilarity, as_recording, Similarity, jaro_winkler, \
    get_similarity_by_name
from token_distance.types.registry import get_if_needed
from token_distance.types.token import WeightedToken, RecordingToken, Token, split_text_by, parse_token
from token_distance.weights import Weight, even_weight, get_weight_by_name


class TokenMatchingAlgorithm(Protocol):
    """
    TokenMatchingAlgorithm is a protocol for classes implementing the token matching algorithm.

    The protocol requires the class to define a __call__ method that takes two string arguments,
    'left' and 'right', and returns a collection of RecordingToken objects.

    Attributes:
        - left (str): The left string for token matching.
        - right (str): The right string for token matching.

    Returns:
        Collection[RecordingToken]: A collection of RecordingToken objects.

    """
    def __call__(self, left: str, right: str) -> Collection[RecordingToken]:
        ...


def match_tokens(
    similarity: Similarity,
    weight: Weight,
    exclusive: bool,
    tokens: Collection[Token],
    candidates: Collection[str]
) -> Collection[RecordingToken]:
    """
    Match Tokens

    Matches tokens against a collection of candidates using similarity and weight.

    Args:
        similarity (Similarity): The similarity function to compare tokens and candidates.
        weight (Weight): The weight function to assign weights to tokens.
        exclusive (bool, optional): If set to True, only one candidate will be matched per token. Defaults to False.
        tokens (Collection[Token]): The collection of tokens to be matched.
        candidates (Collection[str]): The collection of candidates to match against.


    Returns:
        Collection[RecordingToken]: The collection of matched tokens with additional recording information.

    """
    remaining_candidates: list[str] = list(candidates)
    recording_sim: RecordingSimilarity = as_recording(similarity)
    weighted_tokens: list[WeightedToken] = sorted(weight(tokens), key=attrgetter('weight'), reverse=True)
    evaluated_tokens: list[RecordingToken] = list()
    for token in weighted_tokens:
        if not remaining_candidates:
            evaluated_tokens.append(RecordingToken.matched_from(similarity(token, ''), None))
        else:
            evaluate: Callable[[str], RecordingToken] = partial(recording_sim, token)
            best: RecordingToken = max(map(evaluate, remaining_candidates))
            evaluated_tokens.append(
                replace(
                    best,
                    matched_token_position=remaining_candidates.index(best.matched_with)
                ) if best.matched_with else best
            )
            if best.matched_with is not None and exclusive:
                remaining_candidates.remove(best.matched_with)
    return evaluated_tokens


@dataclass
class MatchConfig:
    """
    The MatchConfig class represents the configuration for matching operations.

    Attributes:
        similarity (Similarity | str): The similarity function or strategy to use for matching. Defaults to "jaro_winkler".
        weight (Weight | str): The weighting strategy to use for matching. Defaults to "even_weight".
        tokenize (Callable[[str], Sequence[str]] | Pattern | str | None): The tokenizer function, regular expression pattern,
            or string separator to use for tokenization. Defaults to None.
        exclusive (bool): Indicates whether the matching should be exclusive, i.e., matching results should only include
            exact matches. Defaults to False.
    """
    similarity: Similarity | str = jaro_winkler
    weight: Weight | str = even_weight
    tokenize: Callable[[str], Sequence[str]] | Pattern | str | None = None
    exclusive: bool = False


def match_from_config(config: MatchConfig) -> TokenMatchingAlgorithm:
    """
    Args:
        config: MatchConfig object that holds the configuration settings for the token matching algorithm.

    Returns:
        TokenMatchingAlgorithm object that performs token matching based on the given configuration.

    """
    base_algorithm: Callable[
        [Collection[Token], Collection[str]], Collection[RecordingToken]
    ] = partial(
        match_tokens,
        get_if_needed(get_similarity_by_name, config.similarity),
        get_if_needed(get_weight_by_name, config.weight),
        config.exclusive
    )

    def matching(left: str, right: str) -> Collection[RecordingToken]:
        """
        Args:
            left: Token text
            right: Text to search in

        Returns:
            Collection of matched tokens

        """
        tokens = list(map(parse_token, split_text_by(left, config.tokenize)))
        candidates = split_text_by(right, config.tokenize)
        return base_algorithm(tokens, candidates)

    return matching


from collections import Counter
from itertools import chain

from jellyfish import jaro_winkler_similarity, levenshtein_distance

from token_distance.types.token import WeightedToken, EvaluatedToken


def jaro_winkler(left: WeightedToken, right: str) -> EvaluatedToken:
    """
    Args:
        left: A WeightedToken object representing the left token.
        right: A string representing the right token.

    Returns:
        An EvaluatedToken object representing the token with the computed Jaro-Winkler similarity added.

    """
    return EvaluatedToken.add_similarity(left, jaro_winkler_similarity(left.token, right))


def levenshtein(left: WeightedToken, right: str) -> EvaluatedToken:
    """
    Args:
        left: A WeightedToken object representing the left token.
        right: A string representing the right token.

    Returns:
        An EvaluatedToken object after calculating and adding the similarity between the left and right tokens using the
        Levenshtein distance algorithm. The similarity is added as a distance value.
    """
    return EvaluatedToken.add_similarity(left, levenshtein_distance(left.token, right), is_distance=True)


def jaccard(left: WeightedToken, right: str) -> EvaluatedToken:
    """
    Calculates the Jaccard distance between a weighted token and a string.

    Args:
        left: WeightedToken object representing the weighted token.
        right: String representing the second token.

    Returns:
        EvaluatedToken object with the calculated similarity.
    """
    left_counts: Counter = Counter(left.token)
    right_counts: Counter = Counter(right)
    difference: int = sum(
        abs(left_counts.get(key, 0) - right_counts.get(key, 0))
        for key in set(chain(left_counts, right_counts))
    )
    return EvaluatedToken.add_similarity(left, difference / max(len(left.token), len(right)), is_distance=True)



from functools import partial, wraps
from operator import attrgetter
from typing import Protocol

from token_distance.similarities.base import levenshtein, jaro_winkler, jaccard
from token_distance.types.registry import AlgorithmRegistry, HasName
from token_distance.types.token import WeightedToken, EvaluatedToken, RecordingToken


class Similarity(HasName, Protocol):
    """

    Protocol `Similarity` defines the contract for calculating similarity between a `WeightedToken` and a string.

    The `Similarity` class implements the `__call__` method, which takes a `WeightedToken` object as `left` and a string
    as `right`, and returns an `EvaluatedToken` object.

    """
    def __call__(self, left: WeightedToken, right: str) -> EvaluatedToken:
        ...


class RecordingSimilarity(Similarity, Protocol):
    """
    Module: RecordingSimilarity

    This module provides the class `RecordingSimilarity` which is responsible for calculating the similarity between a
    `WeightedToken` and a `str` recording the match.
    """
    def __call__(self, left: WeightedToken, right: str) -> RecordingToken:
        ...


def as_recording(similarity: Similarity) -> RecordingSimilarity:
    """
    Creates a RecordingSimilarity from an ordinary Similarity.

    Args:
        similarity: The similarity function to be wrapped.

    Returns:
        A RecordingSimilarity object that wraps around the similarity function and returns RecordingToken objects.
    """
    @wraps(similarity)
    def _wrapped(left: WeightedToken, right: str) -> RecordingToken:
        return RecordingToken.matched_from(similarity(left, right), right)
    return _wrapped


_KNOWN_SIMILARITIES: AlgorithmRegistry[Similarity] = AlgorithmRegistry[Similarity](
    levenshtein, jaro_winkler, jaccard
)

get_supported_similarities = partial(attrgetter('supported'), _KNOWN_SIMILARITIES)
get_similarity_by_name = _KNOWN_SIMILARITIES.__getitem__





from collections.abc import Callable, Sequence, Collection
from dataclasses import dataclass
from re import Pattern
from typing import NewType

Token = NewType('Token', str)


def parse_token(value: str) -> Token:
    return Token(value)

@dataclass(frozen=True)
class WeightedToken:
    """
    A class representing a weighted token.

    Attributes:
        token (Token): The token object.
        weight (float): The weight of the token.

    Raises:
        AttributeError: If the weight is not a positive number.

    Example usage:
        token = Token(value='apple')
        weighted_token = WeightedToken(token=token, weight=0.5)
    """
    token: Token
    weight: float

    def __post_init__(self):
        if self.weight <= 0:
            raise AttributeError('Weights must be positive numbers.')

    def __str__(self) -> str:
        return str(self.token)


@dataclass(frozen=True)
class EvaluatedToken(WeightedToken):
    """
    Data class representing an evaluated token.

    Attributes:
        similarity (float): The similarity value of the token.
        is_distance (bool): Flag indicating whether the similarity represents a distance.
    """
    similarity: float
    is_distance: bool = False

    def __lt__(self, other) -> bool:
        if self.is_distance:
            return other.similarity < self.similarity
        else:
            return self.similarity < other.similarity

    def __float__(self) -> float:
        return float(self.similarity * self.weight)

    @classmethod
    def add_similarity(cls, token: WeightedToken, similarity: float, *, is_distance: bool = False) -> 'EvaluatedToken':
        """Adds similarity to a weighted token and returns an evaluated token.

        Args:
            token: A WeightedToken object representing the token.
            similarity: A float representing the similarity value to be added.
            is_distance: An optional boolean indicating whether the similarity value is a distance measure
                         (default is False).

        Returns:
            An evaluated token object with the updated similarity value.
        """
        return cls(token.token, token.weight, similarity, is_distance)


@dataclass(frozen=True)
class RecordingToken(EvaluatedToken):
    """
    The `RecordingToken` class is a subclass of `EvaluatedToken` and represents a token that has been matched with a
    specific value.

    Attributes:
        token (Token): The token object.
        weight (float): The weight of the token.
        similarity (float): The similarity value of the token.
        is_distance (bool): Flag indicating whether the similarity represents a distance.
        matched_with (str | None): The value the token has been matched with.
        matched_token_position (int | None): The position of the matched token in the recording.
    """
    matched_with: str | None = None
    matched_token_position: int | None = None

    @classmethod
    def matched_from(cls, token: EvaluatedToken, match: str | None) -> 'RecordingToken':
        """
        Args:
            token: An instance of the EvaluatedToken class representing the token.
            match: A string representing the match value.

        Returns:
            An instance of the RecordingToken class.

        """
        return RecordingToken(token.token, token.weight, token.similarity, token.is_distance, match)



def split_text_by(
    text: str,
    by: Callable[[str], Sequence[str]] | Pattern | str | None = None
) -> Collection[str]:
    """
    Splits the given text by a specific delimiter.

    Args:
        text: A string representing the text to be split.
        by: A delimiter for splitting the text. It can be a callable function, a regular expression pattern,
            a string, or None. If None is provided, the text is split using whitespace.

    Returns:
        A collection of strings obtained by splitting the text according to the specified delimiter.

    Raises:
        ValueError: If the 'by' argument is not a valid type.

    Note:
        The 'by' argument can be one of the following types:
            - Callable: A function that takes a string as input and returns a sequence of strings.
            - Pattern: A regular expression pattern object.
            - str: A string representing the delimiter to split the text.
            - None: If None is provided, the text is split using whitespace.
    """
    if by is None:
        return text.split()
    if isinstance(by, Pattern):
        return by.split(text)
    if isinstance(by, str):
        return text.split(by)
    if callable(by):
        return by(text)
    else:
        raise ValueError("Invalid 'by' argument")


from collections.abc import Collection
from functools import partial
from operator import attrgetter
from typing import Protocol

from token_distance.types.registry import AlgorithmRegistry, HasName
from token_distance.types.token import Token, WeightedToken
from token_distance.weights.common import even_weight, length_based_weight


class Weight(HasName, Protocol):
    """ An interface representing a weight calculator for tokens. """
    def __call__(self, tokens: Collection[Token]) -> Collection[WeightedToken]:
        ...


_KNOWN_WEIGHTS: AlgorithmRegistry[Weight] = AlgorithmRegistry[Weight](
    even_weight, length_based_weight
)

get_supported_weights = partial(attrgetter('supported'), _KNOWN_WEIGHTS)
get_weight_by_name = _KNOWN_WEIGHTS.__getitem__




from argparse import ArgumentParser
from errno import EINVAL, ENOENT, EPERM
from pathlib import Path
from re import Pattern, compile as regex, error as RegexError
from sys import stderr, exit

from token_distance.algorithms.token_distance import from_config, Config
from token_distance.means import supported_means
from token_distance.similarities import get_supported_similarities
from token_distance.weights import get_supported_weights


def compare_files() -> None:
    """
    Calculates token distance between two files.
    """
    parser: ArgumentParser = ArgumentParser('Calculates token distance between two files.')
    parser.add_argument('token', type=Path, help='Path to the first file containing the tokens to search.')
    parser.add_argument('data', type=Path, help='Path to the second file to be searched in.')
    parser.add_argument(
        '--similarity', default='jaro_winkler', choices=set(get_supported_similarities()),
        help='Similarity to calculate token distance (JaroWinkler)'
    )
    parser.add_argument(
        '--weight', default='even_weight', choices=set(get_supported_weights()),
        help='Weight strategy to calculate token weights (Even)'
    )
    parser.add_argument(
        '--mean', default='average', choices=set(supported_means()),
        help='Aggregating function to calculate distance form token distance (Average)'
    )
    parser.add_argument(
        '--exclusive', type=bool, default=False,
        help='Matched words cannot be matched again (False)'
    )
    parser.add_argument('--tokenize-by', default=None, help='Token to split text (Whitespace)')
    parser.add_argument('--regex', default=False, help='Interpret tokenize-by as regex (False)')
    parser.add_argument('--token-encoding', type=str, default='utf-8', help='Token file encoding (UTF8)')
    parser.add_argument('--search-encoding', type=str, default='utf-8', help='Search file encoding (UTF8)')
    parser.add_argument('--errors', type=str, default='replace', help='Encoding error handling. (replace)')
    args = parser.parse_args()
    try:
        with args.token.open(encoding=args.token_encoding, errors=args.errors) as src:
            token_content: str = src.read()
        with args.data.open(encoding=args.search_encoding, errors=args.errors) as src:
            search_content: str = src.read()

        tokenize_by: str | Pattern | None = args.tokenize_by
        if args.regex:
            tokenize_by = regex(tokenize_by) if tokenize_by else None
            if tokenize_by is None:
                print("Warning: Ignoring --regex flag as --tokenize-by is not set", file=stderr)

        distance: float = from_config(
            Config(args.similarity, args.mean, args.weight, tokenize_by, args.exclusive)
        )(token_content, search_content)
        print(f"Token distance: {distance}")
    except RegexError as error:
        print(f"Error parsing regular expression: {error}", file=stderr)
        exit(EINVAL)
    except FileNotFoundError as error:
        print(f"File does not exist: {error}", file=stderr)
        exit(ENOENT)
    except PermissionError as error:
        print(f"Unsufficient rights to open files: {error}", file=stderr)
        exit(EPERM)
    except Exception as error:
        print(f'Error: {error}', file=stderr)
        exit(1)









from collections.abc import Collection
from functools import partial
from operator import attrgetter
from typing import Protocol

from token_distance.means.base import average, harmonic, geometric
from token_distance.types.registry import AlgorithmRegistry, HasName
from token_distance.types.token import EvaluatedToken


class Mean(HasName, Protocol):
    """
    Calculate the mean value of a collection of evaluated tokens.

    This class defines a protocol for calculating the mean value of a collection of evaluated tokens.
    The `__call__` method takes a collection of evaluated tokens as input and returns the mean value as a float.

    Example usage:
        mean = Mean()
        values = [0.5, 1.5, 2.5]
        result = mean(values) # returns 1.5
    """
    def __call__(self, values: Collection[EvaluatedToken]) -> float:
        ...


_KNOWN_MEANS: AlgorithmRegistry[Mean] = AlgorithmRegistry[Mean](
    average, harmonic, geometric
)


supported_means = partial(attrgetter('supported'), _KNOWN_MEANS)
get_mean_by_name = _KNOWN_MEANS.__getitem__




from token_distance.algorithms.token_distance import token_distance, exclusive_token_distance, from_config, Config
from token_distance.algorithms.token_matching import match_tokens, match_from_config, MatchConfig, RecordingToken


from collections.abc import Collection

from token_distance.types.token import Token, WeightedToken


def even_weight(tokens: Collection[Token]) -> Collection[WeightedToken]:
    """
    Assigns each token a weight of 1.

    Args:
        tokens: A collection of tokens.

    Returns:
        A collection of WeightedToken objects, where each token in the tokens collection is paired with a weight of 1.0.
    """
    return tuple(WeightedToken(token, 1.0) for token in tokens)


def length_based_weight(tokens: Collection[Token]) -> Collection[WeightedToken]:
    """
    Assigns each token it's length as weight

    Args:
        tokens: A collection of tokens.

    Returns:
        Collection[WeightedToken]: A collection of WeightedToken objects, where each object contains a token from the input collection with its corresponding length as the weight.
    """
    return tuple(WeightedToken(token, len(token)) for token in tokens)


from collections.abc import Collection
from functools import reduce
from operator import attrgetter, mul

from token_distance.types.token import EvaluatedToken


def average(values: Collection[EvaluatedToken]) -> float:
    """
    Calculates the average of a collection of evaluated tokens.

    Args:
        values: A collection of evaluated tokens.

    Returns:
        The average value as a float.

    """
    if not values:
        return 0.0
    return sum(map(float, values)) / sum(map(attrgetter('weight'), values))


def geometric(values: Collection[EvaluatedToken]) -> float:
    """
    Calculate the geometric mean of a collection of evaluated tokens.

    Args:
        values (Collection[EvaluatedToken]): A collection of evaluated tokens.

    Returns:
        float: The geometric mean of the given collection of evaluated tokens. If the collection is empty, 0.0 is
        returned.

    Note:
        The geometric mean is calculated by multiplying all the token values, converting them to floats, and then taking
        the nth root, where n is the sum of the weights of the tokens.
    """
    if not values:
        return 0.0
    weights: float = sum(map(attrgetter('weight'), values))
    return reduce(mul, (value.similarity ** value.weight for value in values)) ** (1.0 / weights)


def harmonic(values: Collection[EvaluatedToken]) -> float:
    """
    Calculates the harmonic mean of a collection of evaluated tokens.

    Args:
        values: A collection of evaluated tokens.

    Returns:
        The harmonic mean as a float.
    """
    if not values:
        return 0.0
    weights: float = sum(map(attrgetter('weight'), values))
    try:
        return weights / sum(token.weight / token.similarity for token in values)
    except ZeroDivisionError:
        return 0.0